package xyz.etztech.orealert.listeners;

import com.google.gson.JsonArray;
import com.google.gson.JsonObject;
import net.md_5.bungee.api.ChatColor;
import net.md_5.bungee.api.chat.BaseComponent;
import net.md_5.bungee.api.chat.ComponentBuilder;
import org.bukkit.Bukkit;
import org.bukkit.Material;
import org.bukkit.configuration.file.FileConfiguration;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.Listener;
import org.bukkit.event.block.BlockBreakEvent;
import xyz.etztech.orealert.Color;
import xyz.etztech.orealert.Lang;
import xyz.etztech.orealert.OreAlert;

import java.io.OutputStream;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.ConcurrentLinkedQueue;

public class BlockBreakListener implements Listener {

    private final OreAlert plugin;
    private final Map<Player, List<BlockEvent>> map = new HashMap<>();
    private final Queue<BlockEvent> queue = new ConcurrentLinkedQueue<>();

    public BlockBreakListener(OreAlert plugin) {
        this.plugin = plugin;
        this.plugin.getServer().getPluginManager().registerEvents(this, plugin);
        this.plugin.getServer().getScheduler().runTaskAsynchronously(this.plugin, this::task);
        this.plugin.getServer().getScheduler().runTaskTimerAsynchronously(this.plugin, this::cleanup, 0,
                20 * 60 * this.plugin.getConfig().getInt("cleanup", 5));
    }

    @EventHandler
    public void onBlockBreak(BlockBreakEvent event) {
        queue.add(new BlockEvent(event.getPlayer(), event.getBlock().getType(), true));

        int radius = this.plugin.getConfig().getInt("radius", 3);
        for (int x = -radius; x < radius; x++) {
            for (int y = -radius; y < radius; y++) {
                for (int z = -radius; z < radius; z++) {
                    queue.add(new BlockEvent(event.getPlayer(), event.getBlock().getRelative(x, y, z).getType(), false));
                }
            }
        }
    }

    private void task() {
        while (this.plugin.isEnabled()) {
            BlockEvent event = this.queue.poll();
            if (event != null) {
                for (String s: this.plugin.getConfig().getConfigurationSection("blocks").getKeys(false)) {
                    if (Lang.getMaterialKey(event.getMaterial()).equals(s)) {
                        addStrike(event);
                        check(event);
                        break;
                    }
                }
            }
        }
    }

    private void addStrike(BlockEvent event) {
        List<BlockEvent> events = map.getOrDefault(event.getPlayer(), new ArrayList<>());
        events.add(event);
        map.put(event.getPlayer(), events);
    }

    private void check(BlockEvent event) {
        FileConfiguration config = this.plugin.getConfig();
        String blockKey = Lang.getMaterialKey(event.getMaterial());
        int start = config.getInt(String.format("blocks.%s.start", blockKey), config.getInt("start", 5));
        int each = config.getInt(String.format("blocks.%s.each", blockKey), config.getInt("each", 5));
        int ping = config.getInt(String.format("blocks.%s.ping", blockKey), config.getInt("ping", 5));
        purge(map.getOrDefault(event.getPlayer(), new ArrayList<>()).iterator());

        int strikes = 0;
        for (BlockEvent e : map.getOrDefault(event.getPlayer(), new ArrayList<>())) {
            if (e.isParent()) {
                strikes++;
            }
        }

        double alert = (double) strikes / start;
        if (alert == 1) {
            alert(event, strikes, true);
        } else if (alert > 1) {
            if (alert % ping == 0) {
                alert(event, strikes, true);
            } else if (alert % each == 0) {
                alert(event, strikes, false);
            }
        }
    }

    private void cleanup() {
        for (Iterator<Player> it = map.keySet().iterator(); it.hasNext(); ) {
            Player player = it.next();
            purge(map.get(player).iterator());
            if (map.get(player).isEmpty()) {
                it.remove();
            }
        }
    }

    private void purge(Iterator<BlockEvent> events) {
        Date now = Calendar.getInstance().getTime();
        int globalPurge = this.plugin.getConfig().getInt("purge", 30);
        while (events.hasNext()) {
            BlockEvent e = events.next();
            int purge = 1000 * 60 * this.plugin.getConfig().getInt(String.format("blocks.%s.purge", Lang.getMaterialKey(e.getMaterial())), globalPurge);
            if (new Date(e.getTime().getTime() + purge).before(now)) {
                events.remove();
            }
        }
    }

    private void alert(BlockEvent event, int strikes, boolean ping) {
        String message = String.format(Lang.ALERT.getMessage(),
                event.getPlayer().getName(), strikes, Lang.getMaterialName(event.getMaterial()));
        String hexColor = this.plugin.getConfig().getString(String.format("blocks.%s.color", Lang.getMaterialKey(event.getMaterial())), "#00ffff");
        BaseComponent[] component = new ComponentBuilder()
                .append(message).color(ChatColor.of(hexColor))
                .create();

        for (Player player : Bukkit.getOnlinePlayers()) {
            if (player.hasPermission("orealert.alert")) {
                player.spigot().sendMessage(component);
            }
        }

        // Webhook
        String globalWebhook = this.plugin.getConfig().getString("webhook", "");
        String webhook = this.plugin.getConfig().getString(String.format("blocks.%s.webhook", Lang.getMaterialKey(event.getMaterial())), globalWebhook);
        if (!"".equals(webhook)) {
            String embed = embed(event, message, ping, hexColor);
            this.plugin.getServer().getScheduler().runTaskAsynchronously(this.plugin, () -> {
                try {
                    sendWebhook(webhook, embed);
                } catch (Exception e) {
                    this.plugin.log(String.format("Could not send webhook: %s", e.getMessage()));
                }
            });
        }
    }

    private String embed(BlockEvent event, String message, boolean ping, String hexColor) {
        JsonObject json = new JsonObject();
        if (ping) {
            json.addProperty("content", "@here");
        }
        JsonArray embeds = new JsonArray();

        JsonObject embed = new JsonObject();
        embed.addProperty("color", Color.hexToInt(hexColor));
        embed.addProperty("description", message);

        JsonObject author = new JsonObject();
        author.addProperty("name", event.getPlayer().getName());
        author.addProperty("icon_url", String.format("https://minotar.net/helm/%s/100.png", event.getPlayer().getName()));
        embed.add("author", author);

        embeds.add(embed);
        json.add("embeds", embeds);

        return json.toString();
    }

    private void sendWebhook(String webhook, String embed) throws Exception {
        URL url = new URL(webhook);
        URLConnection con = url.openConnection();
        HttpURLConnection http = (HttpURLConnection) con;
        http.setRequestMethod("POST");
        http.setDoOutput(true);

        byte[] out = embed.getBytes(StandardCharsets.UTF_8);
        int length = out.length;
        http.setFixedLengthStreamingMode(length);
        http.setRequestProperty("Content-Type", "application/json; charset=UTF-8");
        http.connect();

        OutputStream os = http.getOutputStream();
        os.write(out);
    }
}

class BlockEvent {
    private final Player player;
    private final Material material;
    private final Boolean parent;
    private final Date time;

    BlockEvent(Player player, Material material, Boolean parent) {
        this.player = player;
        this.material = material;
        this.parent = parent;
        this.time = Calendar.getInstance().getTime();
    }

    public Player getPlayer() {
        return player;
    }

    public Material getMaterial() {
        return material;
    }

    public Boolean isParent() {
        return parent;
    }

    public Date getTime() {
        return time;
    }
}
